package vector

import (
	"fmt"
	"math/rand"
	"strconv"
	"strings"
	"testing"

	"github.com/stretchr/testify/assert"

	"github.com/Fantom-foundation/go-lachesis/hash"
	"github.com/Fantom-foundation/go-lachesis/inter"
	"github.com/Fantom-foundation/go-lachesis/inter/idx"
	"github.com/Fantom-foundation/go-lachesis/inter/pos"
	"github.com/Fantom-foundation/go-lachesis/kvdb/memorydb"
	"github.com/Fantom-foundation/go-lachesis/logger"
)

func BenchmarkIndex_ForklessCause(b *testing.B) {
	for i := 0; i < b.N; i++ {
		b.StopTimer()
		benchForklessCauseMain(b, &i)
	}
}

func benchForklessCauseMain(b *testing.B, idx *int) {
	benchForklessCauseProcess(b, `
a0_1(3)  b0_1     c0_1     d0_1     e0_1     f0_1     g0_1     h0_1     i0_1     j0_1     k0_1     l0_1     m0_1     p0_1     q0_1
║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║
╠─────── b1_2     ║        ╠─────── e1_2     ║        ╠─────── h1_2     ║        ║        ║        ╠─────── m1_2     ║        ║
║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║
║        ╠─────── c1_3     ║        ╠─────── f1_3     ║        ║        ╠─────── j1_3     ╠─────── l1_3     ╠─────── p1_3     ║
║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║
║        ║        ╠─────── d1_4     ║        ╠─────── g1_4     ╠─────── i1_4     ╠─────── k1_4     ║        ║        ╠─────── q1_4
║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║        ║
`, idx)
}

func benchForklessCauseProcess(b *testing.B, dag string, idx *int) {
	logger.SetTestMode(b)

	nodes, _, _ := inter.ASCIIschemeToDAG(dag)
	validators := pos.EqualStakeValidators(nodes, 1)

	events := make(map[hash.Event]*inter.EventHeaderData)
	getEvent := func(id hash.Event) *inter.EventHeaderData {
		return events[id]
	}

	vi := NewIndex(DefaultIndexConfig(), validators, memorydb.New(), getEvent)

	_, _, named := inter.ASCIIschemeForEach(dag, inter.ForEachEvent{
		Process: func(e *inter.Event, name string) {
			events[e.Hash()] = &e.EventHeaderData
			vi.Add(&e.EventHeaderData)
			vi.Flush()
		},
	})

	// check
	b.StartTimer()
	for _, ev := range named {
		for _, by := range named {
			who := by.Hash()
			whom := ev.Hash()
			vi.ForklessCause(who, whom)
			if *idx > b.N {
				b.StopTimer()
				return
			}
			*idx++
		}
	}
	b.StopTimer()
}

func TestForklessCausedClassic(t *testing.T) {

	t.Run("step 3", func(t *testing.T) {
		testForklessCaused(t, `
a0_1(3)  b0_1     c0_1
║        ║        ║
╠─────── b1_2     ║
║        ║        ║
║        ╠─────── c1_3
║        ║        ║
`)
	})

	t.Run("step 4", func(t *testing.T) {
		testForklessCaused(t, `
a0_1(3)  b0_1     c0_1
║        ║        ║
╠─────── b1_2     ║
║        ║        ║
║        ╠─────── c1_3
║        ║        ║
║        b2_4 ────╣
║        ║        ║
`)
	})

	t.Run("step 5", func(t *testing.T) {
		testForklessCaused(t, `
a0_1(3)  b0_1(5)  c0_1(5)
║        ║        ║
╠─────── b1_2(5)  ║
║        ║        ║
║        ╠─────── c1_3(5)
║        ║        ║
║        b2_4 ────╣
║        ║        ║
a1_5 ────╣        ║
║        ║        ║
`)
	})

}

// testForklessCaused uses event name agreement:
//  "<name>_<level>[(by-level)]",
// where by-level means that event is forkless seen by all event with level >= by-level.
func testForklessCaused(t *testing.T, dag string) {
	logger.SetTestMode(t)
	assertar := assert.New(t)

	nodes, _, _ := inter.ASCIIschemeToDAG(dag)
	validators := pos.EqualStakeValidators(nodes, 1)

	events := make(map[hash.Event]*inter.EventHeaderData)
	getEvent := func(id hash.Event) *inter.EventHeaderData {
		return events[id]
	}

	vi := NewIndex(DefaultIndexConfig(), validators, memorydb.New(), getEvent)

	_, _, named := inter.ASCIIschemeForEach(dag, inter.ForEachEvent{
		Process: func(e *inter.Event, name string) {
			events[e.Hash()] = &e.EventHeaderData
			vi.Add(&e.EventHeaderData)
			vi.Flush()
		},
	})

	// check
	for dsc, ev := range named {
		_, bylevel := decode(dsc)
		for dsc, by := range named {
			level, _ := decode(dsc)

			who := by.Hash()
			whom := ev.Hash()
			if !assertar.Equal(
				bylevel > 0 && bylevel <= level,
				vi.ForklessCause(who, whom),
				fmt.Sprintf("%s forkless sees %s", who.String(), whom.String()),
			) {
				return
			}
		}
	}
}

func decode(dsc string) (level, bylevel int64) {
	var err error

	s := strings.Split(dsc, "_")
	s = strings.Split(s[1], "(")

	level, err = strconv.ParseInt(s[0], 10, 32)
	if err != nil {
		panic(err)
	}

	if len(s) < 2 {
		return
	}

	bylevel, err = strconv.ParseInt(strings.TrimRight(s[1], ")"), 10, 32)
	if err != nil {
		panic(err)
	}

	return
}

func TestForklessCausedRandom(t *testing.T) {
	assertar := assert.New(t)

	// generated by codegen4ForklessCausedStability()
	dag := `
 a000    
 ║        ║       
 ╠═══════ b000    
 ║        ║        ║       
 ║        ║        c000    
 ║        ║        ║        ║       
 ╠═══════─╫─══════─╫─══════ d000    
 ║        ║        ║        ║       
 a001════─╫─══════─╫─═══════╣       
 ║        ║        ║        ║       
 ╠═══════ b001     ║        ║       
 ║        ║        ║        ║       
 ║        ╠═══════ c001     ║       
 ║        ║        ║        ║       
 ║        b002═════╣        ║       
 ║        ║        ║        ║       
 ║        ║        ╠═══════ d001    
 ║        ║        ║        ║       
 a002════─╫─══════─╫─═══════╣       
 ║        ║        ║        ║       
 a003═════╣        ║        ║       
 ║        ║        ║        ║       
 ╠═══════ b003     ║        ║       
 ║        ║        ║        ║       
 ║        ║        c002═════╣       
 ║        ║║       ║        ║       
 ║        ║╚══════─╫─══════ d002    
 ║        ║        ║        ║       
 ║        ║        c003═════╣       
 ║        ║        ║        ║       
 ║        ║        ╠═══════ d003    
 ║        ║        ║        ║       
 a004════─╫─══════─╫─═══════╣       
 ║        ║        ║        ║       
 ╠═══════ b004     ║        ║       
 ║        ║        ║        ║       
 ║        ║        c004═════╣       
 ║        ║        ║        ║       
 a005════─╫─═══════╣        ║       
 ║        ║        ║        ║       
 ╠═══════ b005     ║        ║       
 ║        ║        ║        ║       
 ║        ╠═══════ c005     ║       
 ║        ║        ║        ║       
 a006════─╫─═══════╣        ║       
 ║        ║        ║║       ║       
 ║        ║        ║╚══════ d004    
 ║        ║        ║        ║       
 ║        ╠═══════─╫─══════ d005    
 ║        ║        ║        ║       
 ║        b006════─╫─═══════╣       
 ║        ║        ║        ║       
 a007═════╣        ║        ║       
 ║        ║        ║        ║       
 ║        ║        c006═════╣       
 ║        ║        ║        ║       
 ║        ╠═══════─╫─══════ d006    
 ║        ║        ║        ║       
 ║        b007════─╫─═══════╣       
 ║        ║        ║        ║       
 ║        ║        c007═════╣       
 ║        ║        ║        ║       
 ╠═══════─╫─══════─╫─══════ d007    
 ║        ║        ║        ║       
 a008════─╫─══════─╫─═══════╣       
 ║        ║        ║        ║       
 ║        b008════─╫─═══════╣       
 ║        ║        ║        ║       
 ║        ╠═══════ c008     ║       
 ║        ║        ║        ║       
 a009════─╫─═══════╣        ║       
 ║        ║        ║        ║       
 ╠═══════ b009     ║        ║       
 ║        ║        ║        ║       
 ║        ║        ╠═══════ d008    
 ║        ║        ║        ║       
 ║        ║        c009═════╣       
 ║        ║        ║        ║       
 ║        b010═════╣        ║       
 ║        ║        ║        ║       
 ║        ║        ╠═══════ d009    
 ║        ║        ║        ║       
 a010════─╫─══════─╫─═══════╣       
 ║        ║        ║        ║       
 ╠═══════─╫─══════ c010     ║       
 ║        ║        ║        ║       
 ║        b011═════╣        ║       
 ║        ║        ║        ║       
 ║        ╠═══════ c011     ║       
 ║        ║║       ║        ║       
 ║        ║╚══════─╫─══════ d010    
 ║        ║        ║        ║       
 a011════─╫─══════─╫─═══════╣       
 ║        ║        ║        ║       
 a012════─╫─═══════╣        ║       
 ║        ║        ║        ║       
 ╠═══════ b012     ║        ║       
 ║        ║        ║        ║       
 ╠═══════─╫─══════ c012     ║       
 ║        ║        ║        ║       
 ║        b013═════╣        ║       
 ║        ║        ║        ║       
 ║        ╠═══════ c013     ║       
 ║        ║3       ║        ║       
 ║        ║╚══════─╫─══════ d011    
 ║        ║        ║║       ║       
 ║        ║        ║╚══════ d012    
 ║        ║        ║        ║       
 a013════─╫─══════─╫─═══════╣       
 ║        ║        ║        ║       
 a014═════╣        ║        ║       
 ║        ║        ║        ║       
 ╠═══════─╫─══════ c014     ║       
 ║        ║        ║║       ║       
 ║        ║        ║╚══════ d013    
 ║        ║        ║        ║       
 ║        b014════─╫─═══════╣       
 ║        ║        ║        ║       
 a015═════╣        ║        ║       
 ║        ║        ║        ║       
 ║        b015═════╣        ║       
 ║        ║        ║        ║       
 ║        ╠═══════ c015     ║       
 ║        ║        ║        ║       
 a016════─╫─═══════╣        ║       
 ║        ║        ║        ║       
 ║        b016═════╣        ║       
 ║        ║        ║        ║       
 ╠═══════─╫─══════ c016     ║       
 ║        ║        ║        ║       
 a017════─╫─═══════╣        ║       
 ║        ║        ║        ║       
 ║        b017═════╣        ║       
 ║        ║        ║        ║       
 a018═════╣        ║        ║       
 ║5       ║        ║        ║       
 ║╚══════─╫─══════─╫─══════ d014    
 ║        ║        ║║       ║       
 ║        ║        ║╚══════ d015    
 ║3       ║        ║        ║       
 ║╚══════─╫─══════─╫─══════ d016    
 ║        ║        ║        ║       
 ║        ║        c017═════╣       
 ║        ║        ║        ║       
 ║        b018═════╣        ║       
 ║        ║        ║        ║       
 ║        ╠═══════ c018     ║       
 ║        ║║       ║        ║       
 ║        ║╚══════─╫─══════ d017    
 ║        ║        ║        ║       
 ║        ║        ╠═══════ d018    
 ║        ║        ║        ║       
 a019════─╫─══════─╫─═══════╣       
 ║        ║        ║        ║       
 ║        b019════─╫─═══════╣       
 ║        ║        ║        ║       
 ╠═══════─╫─══════ c019     ║       
 ║        ║        ║        ║       
 ║        ║        ╠═══════ d019    
`
	relations := map[string]map[string]struct{}{
		"a000": map[string]struct{}{},
		"a001": map[string]struct{}{},
		"a002": map[string]struct{}{"a000": {}, "a001": {}, "b000": {}, "b001": {}, "c000": {}, "c001": {}, "d000": {}},
		"a003": map[string]struct{}{"a000": {}, "a001": {}, "b000": {}, "b001": {}, "c000": {}, "c001": {}, "d000": {}},
		"a004": map[string]struct{}{"a000": {}, "a001": {}, "b000": {}, "b001": {}, "b002": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "d000": {}, "d001": {}, "d002": {}},
		"a005": map[string]struct{}{"a000": {}, "a001": {}, "b000": {}, "b001": {}, "b002": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}},
		"a006": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}},
		"a007": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}},
		"a008": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}},
		"a009": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}},
		"a010": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}},
		"a011": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}},
		"a012": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}},
		"a013": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}},
		"a014": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}},
		"a015": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}},
		"a016": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}},
		"a017": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}},
		"a018": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "a015": {}, "a016": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "c015": {}, "c016": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}},
		"a019": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "a015": {}, "a016": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "b016": {}, "b017": {}, "b018": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "c015": {}, "c016": {}, "c017": {}, "c018": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}, "d014": {}, "d015": {}, "d016": {}},
		"b000": map[string]struct{}{},
		"b001": map[string]struct{}{"a000": {}, "d000": {}},
		"b002": map[string]struct{}{"a000": {}, "a001": {}, "d000": {}},
		"b003": map[string]struct{}{"a000": {}, "a001": {}, "b000": {}, "b001": {}, "c000": {}, "c001": {}, "d000": {}, "d001": {}},
		"b004": map[string]struct{}{"a000": {}, "a001": {}, "b000": {}, "b001": {}, "b002": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}},
		"b005": map[string]struct{}{"a000": {}, "a001": {}, "b000": {}, "b001": {}, "b002": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}},
		"b006": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "b000": {}, "b001": {}, "b002": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}},
		"b007": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "b000": {}, "b001": {}, "b002": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}},
		"b008": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}},
		"b009": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}},
		"b010": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}},
		"b011": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}},
		"b012": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}},
		"b013": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}},
		"b014": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}},
		"b015": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}},
		"b016": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}},
		"b017": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "a015": {}, "a016": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "c015": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}},
		"b018": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "a015": {}, "a016": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "c015": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}, "d014": {}, "d015": {}, "d016": {}},
		"b019": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "a015": {}, "a016": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "b016": {}, "b017": {}, "b018": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "c015": {}, "c016": {}, "c017": {}, "c018": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}, "d014": {}, "d015": {}, "d016": {}},
		"c000": map[string]struct{}{},
		"c001": map[string]struct{}{"a000": {}, "a001": {}, "d000": {}},
		"c002": map[string]struct{}{"a000": {}, "a001": {}, "b000": {}, "b001": {}, "d000": {}},
		"c003": map[string]struct{}{"a000": {}, "a001": {}, "b000": {}, "b001": {}, "b002": {}, "c000": {}, "c001": {}, "d000": {}},
		"c004": map[string]struct{}{"a000": {}, "a001": {}, "b000": {}, "b001": {}, "b002": {}, "c000": {}, "c001": {}, "d000": {}},
		"c005": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "b000": {}, "b001": {}, "b002": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}},
		"c006": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}},
		"c007": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}},
		"c008": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}},
		"c009": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}},
		"c010": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}},
		"c011": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}},
		"c012": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}},
		"c013": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}},
		"c014": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}},
		"c015": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}},
		"c016": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}},
		"c017": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "a015": {}, "a016": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "c015": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}},
		"c018": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "a015": {}, "a016": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "c015": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}, "d014": {}, "d015": {}, "d016": {}},
		"c019": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "a015": {}, "a016": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "b016": {}, "b017": {}, "b018": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "c015": {}, "c016": {}, "c017": {}, "c018": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}, "d014": {}, "d015": {}, "d016": {}, "d017": {}, "d018": {}},
		"d000": map[string]struct{}{},
		"d001": map[string]struct{}{"a000": {}, "a001": {}, "b000": {}, "b001": {}, "d000": {}},
		"d002": map[string]struct{}{"a000": {}, "a001": {}, "b000": {}, "b001": {}, "c000": {}, "c001": {}, "d000": {}},
		"d003": map[string]struct{}{"a000": {}, "a001": {}, "b000": {}, "b001": {}, "b002": {}, "c000": {}, "c001": {}, "d000": {}},
		"d004": map[string]struct{}{"a000": {}, "a001": {}, "b000": {}, "b001": {}, "b002": {}, "c000": {}, "c001": {}, "d000": {}},
		"d005": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "b000": {}, "b001": {}, "b002": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}},
		"d006": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "b000": {}, "b001": {}, "b002": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}},
		"d007": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}},
		"d008": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}},
		"d009": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}},
		"d010": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}},
		"d011": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}},
		"d012": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}},
		"d013": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}},
		"d014": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}},
		"d015": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}},
		"d016": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "c015": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}},
		"d017": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "a015": {}, "a016": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "c015": {}, "c016": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}},
		"d018": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "a015": {}, "a016": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "b016": {}, "b017": {}, "b018": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "c015": {}, "c016": {}, "c017": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}, "d014": {}, "d015": {}, "d016": {}},
		"d019": map[string]struct{}{"a000": {}, "a001": {}, "a002": {}, "a003": {}, "a004": {}, "a005": {}, "a006": {}, "a007": {}, "a008": {}, "a009": {}, "a010": {}, "a011": {}, "a012": {}, "a013": {}, "a014": {}, "a015": {}, "a016": {}, "a017": {}, "a018": {}, "a019": {}, "b000": {}, "b001": {}, "b002": {}, "b003": {}, "b004": {}, "b005": {}, "b006": {}, "b007": {}, "b008": {}, "b009": {}, "b010": {}, "b011": {}, "b012": {}, "b013": {}, "b014": {}, "b015": {}, "b016": {}, "b017": {}, "b018": {}, "c000": {}, "c001": {}, "c002": {}, "c003": {}, "c004": {}, "c005": {}, "c006": {}, "c007": {}, "c008": {}, "c009": {}, "c010": {}, "c011": {}, "c012": {}, "c013": {}, "c014": {}, "c015": {}, "c016": {}, "c017": {}, "c018": {}, "d000": {}, "d001": {}, "d002": {}, "d003": {}, "d004": {}, "d005": {}, "d006": {}, "d007": {}, "d008": {}, "d009": {}, "d010": {}, "d011": {}, "d012": {}, "d013": {}, "d014": {}, "d015": {}, "d016": {}, "d017": {}, "d018": {}},
	}

	ordered := make([]*inter.Event, 0)
	nodes, _, named := inter.ASCIIschemeForEach(dag, inter.ForEachEvent{
		Process: func(e *inter.Event, name string) {
			ordered = append(ordered, e)
		},
	})

	validators := pos.EqualStakeValidators(nodes, 1)

	events := make(map[hash.Event]*inter.EventHeaderData)
	getEvent := func(id hash.Event) *inter.EventHeaderData {
		return events[id]
	}

	vi := NewIndex(DefaultIndexConfig(), validators, memorydb.New(), getEvent)

	// push
	for _, e := range ordered {
		events[e.Hash()] = &e.EventHeaderData
		vi.Add(&e.EventHeaderData)
		vi.Flush()
	}

	// check
	for e1name, e1 := range named {
		for e2name, e2 := range named {
			_, expect := relations[e1name][e2name]
			if !assertar.Equal(
				expect,
				vi.ForklessCause(e1.Hash(), e2.Hash()),
				fmt.Sprintf("%s forkless sees %s", e1.Hash(), e2.Hash()),
			) {
				return
			}
		}
	}
}

type eventSlot struct {
	seq     idx.Event
	creator idx.StakerID
}

// naive implementation of fork detection, O(n)
func testForksDetected(vi *Index, head *inter.EventHeaderData) (cheaters map[idx.StakerID]bool, err error) {
	cheaters = map[idx.StakerID]bool{}
	visited := hash.EventsSet{}
	detected := map[eventSlot]int{}
	onWalk := func(id hash.Event) (godeeper bool) {
		// ensure visited once
		if visited.Contains(id) {
			return false
		}
		visited.Add(id)

		e := vi.getEvent(id)
		slot := eventSlot{
			seq:     e.Seq,
			creator: e.Creator,
		}
		detected[slot]++
		return true
	}
	onWalk(head.Hash())
	err = vi.dfsSubgraph(head, onWalk)
	for s, count := range detected {
		if count > 1 {
			cheaters[s.creator] = true
		}
	}
	return cheaters, err
}

func TestRandomForksSanity(t *testing.T) {
	nodes := inter.GenNodes(8)
	cheaters := []idx.StakerID{nodes[0], nodes[1], nodes[2]}

	validatorsBuilder := pos.NewBuilder()
	for _, peer := range nodes {
		validatorsBuilder.Set(peer, pos.Stake(1))
	}

	validatorsBuilder.Set(cheaters[0], pos.Stake(2))
	validatorsBuilder.Set(nodes[3], pos.Stake(2))
	validatorsBuilder.Set(nodes[4], pos.Stake(3))
	validators := validatorsBuilder.Build()

	processed := make(map[hash.Event]*inter.EventHeaderData)
	getEvent := func(id hash.Event) *inter.EventHeaderData {
		return processed[id]
	}

	vi := NewIndex(DefaultIndexConfig(), validators, memorydb.New(), getEvent)

	// Many forks from each node in large graph, so probability of not seeing a fork is negligible
	events := inter.ForEachRandFork(nodes, cheaters, 300, 4, 30, nil, inter.ForEachEvent{
		Process: func(e *inter.Event, name string) {
			if _, ok := processed[e.Hash()]; ok {
				return
			}
			processed[e.Hash()] = &e.EventHeaderData
			vi.Add(&e.EventHeaderData)
		},
	})

	vi.Flush()
	vi.DropNotFlushed() // doesn't drop anything, because everything is flushed

	// quick sanity check. all the nodes should see that cheaters have a fork, and honest nodes don't have forks
	assertar := assert.New(t)
	idxs := validatorsBuilder.Build().Idxs()
	for _, node := range nodes {
		ee := events[node]
		highestBefore := vi.GetHighestBeforeSeq(ee[len(ee)-1].Hash())
		for n, cheater := range nodes {
			branchSeq := highestBefore.Get(idxs[cheater])
			isCheater := n < len(cheaters)
			assertar.Equal(isCheater, branchSeq.IsForkDetected(), cheater)
			if isCheater {
				assertar.Equal(idx.Event(0), branchSeq.Seq, cheater)
			} else {
				assertar.NotEqual(idx.Event(0), branchSeq.Seq, cheater)
			}
		}
	}
}

func TestRandomForks(t *testing.T) {
	for i, test := range []struct {
		nodesNum      int
		cheatersNum   int
		eventsNum     int
		forksNum      int
		parentsNum    int
		reorderChecks int
	}{
		{
			nodesNum:      1,
			parentsNum:    1,
			cheatersNum:   1,
			eventsNum:     10,
			forksNum:      3,
			reorderChecks: 30,
		},
		{
			nodesNum:      2,
			parentsNum:    1,
			cheatersNum:   1,
			eventsNum:     10,
			forksNum:      3,
			reorderChecks: 20,
		},
		{
			nodesNum:      2,
			parentsNum:    2,
			cheatersNum:   2,
			eventsNum:     10,
			forksNum:      20,
			reorderChecks: 5,
		},
		{
			nodesNum:      10,
			parentsNum:    4,
			cheatersNum:   1,
			eventsNum:     10,
			forksNum:      3,
			reorderChecks: 5,
		},
		{
			nodesNum:      10,
			parentsNum:    4,
			cheatersNum:   10,
			eventsNum:     10,
			forksNum:      3,
			reorderChecks: 3,
		},
		{
			nodesNum:      20,
			parentsNum:    4,
			cheatersNum:   10,
			eventsNum:     5,
			forksNum:      2,
			reorderChecks: 3,
		},
		{
			nodesNum:      40,
			parentsNum:    4,
			cheatersNum:   10,
			eventsNum:     3,
			forksNum:      1,
			reorderChecks: 2,
		},
		{
			nodesNum:      5,
			parentsNum:    4,
			cheatersNum:   2,
			eventsNum:     30,
			forksNum:      30,
			reorderChecks: 2,
		},
	} {
		t.Run(fmt.Sprintf("Test #%d", i), func(t *testing.T) {
			testTime := 100

			r := rand.New(rand.NewSource(int64(i)))

			nodes := inter.GenNodes(test.nodesNum)
			cheaters := nodes[:test.cheatersNum]

			validators := pos.EqualStakeValidators(nodes, 1)

			processedArr := inter.Events{}
			processed := make(map[hash.Event]*inter.EventHeaderData)
			getEvent := func(id hash.Event) *inter.EventHeaderData {
				return processed[id]
			}

			vi := NewIndex(DefaultIndexConfig(), validators, memorydb.New(), getEvent)

			_ = inter.ForEachRandFork(nodes, cheaters, test.eventsNum, test.parentsNum, test.forksNum, r, inter.ForEachEvent{
				Process: func(e *inter.Event, name string) {
					if _, ok := processed[e.Hash()]; ok {
						return
					}
					processed[e.Hash()] = &e.EventHeaderData
					processedArr = append(processedArr, e)
					vi.Add(&e.EventHeaderData)
				},
				Build: func(e *inter.Event, name string) *inter.Event {
					e.ClaimedTime = inter.Timestamp(r.Intn(testTime))
					return e
				},
			})

			assertar := assert.New(t)
			idxs := validators.Idxs()
			// check that fork observing is identical to naive version
			for _, e := range processed {
				highestBefore := vi.GetHighestBeforeSeq(e.Hash())
				expectedCheaters, err := testForksDetected(vi, e)
				assertar.NoError(err)

				for _, cheater := range nodes {
					expectedCheater := expectedCheaters[cheater]
					branchSeq := highestBefore.Get(idxs[cheater])
					assertar.Equal(expectedCheater, branchSeq.IsForkDetected(), e.String())
					if expectedCheater {
						assertar.Equal(idx.Event(0), branchSeq.Seq, e.String())
					}
				}
			}

			// memorize results of ForklessCause and MedianTime
			forklessCauseMap := map[kv]bool{}
			medianTimeMap := map[hash.Event]inter.Timestamp{}
			for _, a := range processedArr {
				medianTimeMap[a.Hash()] = vi.MedianTime(a.Hash(), inter.Timestamp(testTime/2))
				for _, b := range processedArr {
					pair := kv{
						a: a.Hash(),
						b: b.Hash(),
					}
					forklessCauseMap[pair] = vi.ForklessCause(a.Hash(), b.Hash())
				}
			}

			vi.DropNotFlushed() // drops everything, because wasn't flushed
			for _, e := range processed {
				assertar.Nil(vi.GetHighestBeforeSeq(e.Hash()))
				assertar.Nil(vi.GetLowestAfterSeq(e.Hash()))
				assertar.Nil(vi.GetHighestBeforeTime(e.Hash()))
			}

			// check that events re-order doesn't change forklessCause result
			for reorderTry := 0; reorderTry < test.reorderChecks; reorderTry++ {
				// re-order events randomly, preserving parents order
				unordered := make(inter.Events, len(processedArr))
				for i, j := range r.Perm(len(processedArr)) {
					unordered[i] = processedArr[j]
				}
				processedArr = unordered.ByParents()

				for _, a := range processedArr {
					vi.Add(&a.EventHeaderData)
				}

				vi.cache.ForklessCause.Purge() // disable cache
				for _, a := range processedArr {
					res := vi.MedianTime(a.Hash(), inter.Timestamp(testTime/2))
					assertar.Equal(medianTimeMap[a.Hash()], res, "%s %d", a.Hash().String(), reorderTry)

					for _, b := range processedArr {
						pair := kv{
							a: a.Hash(),
							b: b.Hash(),
						}
						res := vi.ForklessCause(a.Hash(), b.Hash())
						assertar.Equal(forklessCauseMap[pair], res, "%s %s %d", a.Hash().String(), b.Hash().String(), reorderTry)
					}
				}
				vi.DropNotFlushed()
			}
		})
	}
}

/*
// codegen4ForklessCausedStability is for test data generation.
func codegen4ForklessCausedStability() {
	peers := inter.GenNodes(4)
	events := inter.GenEventsByNode(peers, 20, 2, nil, nil, nil)

	validators := make(pos.Validators, len(peers))
	for _, peer := range peers {
		validators.Set(peer, pos.Stake(1))
	}
	vi := NewIndex(validators, memorydb.New())

	processed := make(map[hash.Event]*inter.Event)
	orderThenProcess := ordering.EventBuffer(ordering.Callback{

		Process: func(e *inter.Event) {
			processed[e.Hash()] = e
			vi.Add(e)
		},

		Drop: func(e *inter.Event, err error) {
			panic(err)
		},

		Exists: func(h hash.Event) *inter.Event {
			return processed[h]
		},
	})

	// push
	dag := inter.Events{}
	for _, ee := range events {
		for _, e := range ee {
			orderThenProcess(e)
		}
		dag = append(dag, ee...)
	}

	// generation
	scheme, err := inter.DAGtoASCIIscheme(dag)
	if err != nil {
		panic(err)
	}

	fmt.Printf("dag := `%s`\n", scheme)
	fmt.Printf("relations := map[string]map[string]struct{}{\n")
	for _, e1 := range dag {
		sees := fmt.Sprintf("\"%s\": map[string]struct{}{", e1.Hash())
		for _, e2 := range dag {
			if vi.ForklessCause(e1.Hash(), e2.Hash()) {
				sees = sees + fmt.Sprintf("\"%s\":{},", e2.Hash())
			}
		}
		sees += "},"
		fmt.Println(sees)
	}
	fmt.Printf("}\n")
}
*/
